{------------------------------------------------------------------------------}
{   MCM Design                                                                 }
{   Marc Christopher Martin                                                    }
{   Australiensvej 34                                                          }
{   2100 Copenhagen OE                                                         }
{   Denmark                                                                    }
{                                                                              }
{   e-mail support@mcm-design.dk                                               }
{                                                                              }
{------------------------------------------------------------------------------}

unit Main;

// This tool kit does not include the mcmTIFF.DLL file, therefore do not
// define INCTIFF.
{ $DEFINE INCTIFF}

interface

uses WinTypes, WinProcs, SysUtils, Classes, Graphics, Forms, Controls, Menus,
     StdCtrls, Dialogs, Buttons, Messages, ExtCtrls, UAbout,
     {$IFNDEF VER90} { JPEG is not included in Delphi 2. }
       JPEG,
     {$ENDIF}
     {$IFDEF INCTIFF}
       UITIFF,
     {$ENDIF}
     { TWAIN version 1.7, 16/32 bit.                                           }
     twain,   twa_apph, twa_glue, twa_acq,
     twa_capt;


type
  TMainForm = class(TForm)
    MainMenu           : TMainMenu;
    Panel1             : TPanel;
    StatusLine         : TPanel;
    File1              : TMenuItem;
    Panel2             : TPanel;
    Window1            : TMenuItem;
    Help1              : TMenuItem;
    N1                 : TMenuItem;
    FileExitItem       : TMenuItem;
    WindowCascadeItem  : TMenuItem;
    WindowTileItem     : TMenuItem;
    WindowArrangeItem  : TMenuItem;
    HelpAboutItem      : TMenuItem;
    Edit1              : TMenuItem;
    CutItem            : TMenuItem;
    CopyItem           : TMenuItem;
    PasteItem          : TMenuItem;
    WindowMinimizeItem : TMenuItem;
    SpeedPanel         : TPanel;
    OpenBtn            : TSpeedButton;
    SaveBtn            : TSpeedButton;
    sbSource           : TSpeedButton;
    sbAcquire          : TSpeedButton;
    CutBtn             : TSpeedButton;
    CopyBtn            : TSpeedButton;
    PasteBtn           : TSpeedButton;
    ExitBtn            : TSpeedButton;
    N2                 : TMenuItem;
    N3                 : TMenuItem;
    Acquire            : TMenuItem;
    Source             : TMenuItem;
    NativeXfer         : TMenuItem;
    FileXfer           : TMenuItem;
    MemoryXfer         : TMenuItem;
    N4                 : TMenuItem;
    AutomaticFeeding   : TMenuItem;
    N5                 : TMenuItem;
    FileShowUI         : TMenuItem;
    FileDisable        : TMenuItem;
    MsgLevel           : TMenuItem;
    MsgNone            : TMenuItem;
    MsgError           : TMenuItem;
    MsgInformation     : TMenuItem;
    MsgFull            : TMenuItem;
    FileOpenItem       : TMenuItem;
    FileSaveItem       : TMenuItem;
    N6                 : TMenuItem;
    OpenDialog         : TOpenDialog;
    SaveDialog         : TSaveDialog;
    procedure FormCreate             (      Sender  : TObject);
    procedure WindowCascadeItemClick (      Sender  : TObject);
    procedure UpdateMenuItems        (      Sender  : TObject);
    procedure WindowTileItemClick    (      Sender  : TObject);
    procedure WindowArrangeItemClick (      Sender  : TObject);
    procedure FileCloseItemClick     (      Sender  : TObject);
    procedure FileOpenItemClick      (      Sender  : TObject);
    procedure FileExitItemClick      (      Sender  : TObject);
    procedure FileSaveItemClick      (      Sender  : TObject);
    procedure FileSaveAsItemClick    (      Sender  : TObject);
    procedure CutItemClick           (      Sender  : TObject);
    procedure CopyItemClick          (      Sender  : TObject);
    procedure PasteItemClick         (      Sender  : TObject);
    procedure WindowMinimizeItemClick(      Sender  : TObject);
    procedure FormDestroy            (      Sender  : TObject);
    procedure SourceClick            (      Sender  : TObject);
    procedure AcquireClick           (      Sender  : TObject);
    procedure NativeXferClick        (      Sender  : TObject);
    procedure FileXferClick          (      Sender  : TObject);
    procedure MemoryXferClick        (      Sender  : TObject);
    procedure MsgNoneClick           (      Sender  : TObject);
    procedure MsgErrorClick          (      Sender  : TObject);
    procedure MsgInformationClick    (      Sender  : TObject);
    procedure MsgFullClick           (      Sender  : TObject);
    procedure AutomaticFeedingClick  (      Sender  : TObject);
    procedure FileShowUIClick        (      Sender  : TObject);
    procedure FileDisableClick       (      Sender  : TObject);
    procedure HelpAboutItemClick(Sender: TObject);

  protected
    procedure PMXferDone             (var   Msg     : TMessage); message pm_xferdone;
    procedure TWAIN_MessageHook;

  private
    { Private declarations }
    procedure CreateMDIChild         (      Name    : string;
                                            hWinBmp : HBitmap;
                                            w, h    : integer);
    procedure ShowHint               (      Sender  : TObject);

  public
    { Public declarations }
    AppIdentity : TW_IDENTITY;
    {$IFNDEF VER90}
      JPEGImage   : TJPEGImage;
    {$ENDIF}
  end;

var MainForm : TMainForm;


implementation

{$R *.DFM}

uses ChildWin;

{$IFNDEF WIN32}
  procedure AHIncr; far; external 'KERNEL' index 114;
{$ENDIF}


procedure TMainForm.FormCreate(Sender : TObject);
begin
  Application.OnHint        := ShowHint;
  Screen.OnActiveFormChange := UpdateMenuItems;

  { TWAIN                                                                      }
  { Set up the information about your application which you want to pass       }
  { to the SM in this call.                                                    }
  {                                                                            }
  { Move all these text strings off to a resource fill for a real              }
  { application. They are here for easier readablilty for the student.         }
  { init to 0, but Source Manager will assign real value                       }

  with AppIdentity
  do begin
    Id               := 0;
    Version.MajorNum := 1;
    Version.MinorNum := 5;
    Version.Language := TWLG_ENG;
    Version.Country  := TWCY_UNITEDKINGDOM;
    {$IFDEF WIN32}
      lstrcpy(Version.Info,  'TWAIN_32 v1.6.1 Application Name');
    {$ELSE}
      lstrcpy(Version.Info,  'TWAIN_16 v1.6.1 Application Name');
    {$ENDIF}

    ProtocolMajor    := TWON_PROTOCOLMAJOR;
    ProtocolMinor    := TWON_PROTOCOLMINOR;
    SupportedGroups  := DG_IMAGE or DG_CONTROL;
    lstrcpy(Manufacturer,  'MCM Design');
    lstrcpy(ProductFamily, 'Image Capture');
    {$IFDEF WIN32}
      lstrcpy(ProductName,   'TWAIN 32 App');
    {$ELSE}
      lstrcpy(ProductName,   'TWAIN 16 App');
    {$ENDIF}

    { pass app particulars to glue code.                                       }
    TWInitialize(@AppIdentity, Handle);
  end;
  SetMessageLevel(ML_ERROR);
  MsgError.Checked := True;
end;


procedure TMainForm.ShowHint(Sender : TObject);
begin
  StatusLine.Caption := Application.Hint;
end; { End TMainForm.ShowHint.                                                 }


procedure TMainForm.CreateMDIChild(Name     : string;
                                   hWinBmp  : HBitmap;
                                   w, h     : integer);
var Child : TMDIChild;
begin
  { create a new MDI child window }
  if (hWinBmp <> 0) or FileExists(Name) or True
  then begin
       Child         := TMDIChild.Create(Application);
       Child.Caption := Name;
       Child.Color   := clWhite;
       if (hWinBmp <> 0)
       then begin
            Child.ClientWidth   := w;
            Child.ClientHeight  := h;
            Child.Bitmap.Width  := w;
            Child.Bitmap.Height := h;
            Child.Bitmap.Handle := hWinBmp;
{            Child.Image1.Picture.Bitmap.Handle := hWinBmp; }
       end
       else begin
            Name := UpperCase(Name);
            if (Pos('.BMP', Name) <> 0)
            then Child.Bitmap.LoadFromFile(Name);
            {$IFNDEF VER90}
              if (Pos('.JPG', Name) <> 0)
              then begin
                   JPEGImage := TJPEGImage.Create;
                   JPEGImage.Performance := jpBestQuality;
                   JPEGImage.LoadFromFile(Name);
                   Child.Bitmap.Width  := JPEGImage.Width;
                   Child.Bitmap.Height := JPEGImage.Height;
                   Child.Bitmap.Canvas.Draw(0, 0, JPEGImage);
                   JPEGImage.Free;
              end;
            {$ENDIF}
            {$IFDEF INCTIFF}
              if (Pos('.TIF', Name) <> 0)
              then LoadTIFFToBitmap(PChar(Name), Child.Bitmap);
            {$ENDIF}
            Child.ClientWidth   := Child.Bitmap.Width;
            Child.ClientHeight  := Child.Bitmap.Height;
       end;
       Child.HorzScrollBar.Range := w;
       Child.VertScrollBar.Range := h;
       Child.Show;
  end;
end; { End TMainForm.CreateMDIChild.                                           }


procedure TMainForm.FileOpenItemClick(Sender : TObject);
begin
  OpenDialog.Filter   := 'Bitmap|*.BMP';
  {$IFNDEF VER90}
    OpenDialog.Filter := OpenDialog.Filter + '|JPEG|*.JPG';
  {$ENDIF}
  {$IFDEF INCTIFF}
    OpenDialog.Filter := OpenDialog.Filter + '|TIFF 5.0|*.TIF';
  {$ENDIF}
  case OpenDialog.FilterIndex of
  1 : OpenDialog.FileName := '*.BMP';
  2 : OpenDialog.FileName := '*.JPG';
  3 : OpenDialog.FileName := '*.TIF';
  end;
  if OpenDialog.Execute
  then CreateMDIChild(Lowercase(OpenDialog.FileName), 0, 0, 0);
end; { End TMainForm.FileOpenItemClick.                                        }


procedure TMainForm.FileCloseItemClick(Sender : TObject);
begin
  if (ActiveMDIChild <> nil)
  then ActiveMDIChild.Close;
end; { End TMainForm.FileCloseItemClick.                                       }


procedure TMainForm.FileSaveItemClick(Sender : TObject);
begin
  { save current file (ActiveMDIChild points to the window) }
  if (ActiveMDIChild <> nil)
  then begin
       SaveDialog.Filter   := 'Bitmap|*.BMP';
       {$IFNDEF VER90}
         SaveDialog.Filter := SaveDialog.Filter + '|JPEG|*.JPG';
       {$ENDIF}
       {$IFDEF INCTIFF}
         SaveDialog.Filter := SaveDialog.Filter + '|TIFF 5.0|*.TIF';
       {$ENDIF}
       case SaveDialog.FilterIndex of
       1 : SaveDialog.FileName := '*.BMP';
       2 : SaveDialog.FileName := '*.JPG';
       3 : SaveDialog.FileName := '*.TIF';
       end;
       TMDIChild(ActiveMDIChild).Caption := UpperCase(TMDIChild(ActiveMDIChild).Caption);
       if (Pos('.BMP', TMDIChild(ActiveMDIChild).Caption) <> 0)
       then begin
            SaveDialog.FilterIndex := 1;
            SaveDialog.FileName := '*.BMP';
       end;
       if (Pos('.JPG', TMDIChild(ActiveMDIChild).Caption) <> 0)
       then begin
            SaveDialog.FilterIndex := 2;
            SaveDialog.FileName := '*.JPG';
       end;
       if (Pos('.TIF', TMDIChild(ActiveMDIChild).Caption) <> 0)
       then begin
            SaveDialog.FilterIndex := 3;
            SaveDialog.FileName := '*.TIF';
       end;

       if SaveDialog.Execute
       then begin
            if (Pos('.', SaveDialog.FileName) > 0)
            then SaveDialog.FileName := Copy(SaveDialog.FileName, 1, Pos('.', SaveDialog.FileName) - 1);
            case SaveDialog.FilterIndex of
            1 : SaveDialog.FileName := SaveDialog.FileName + '.BMP';
            2 : SaveDialog.FileName := SaveDialog.FileName + '.JPG';
            3 : SaveDialog.FileName := SaveDialog.FileName + '.TIF';
            end;
            TMDIChild(ActiveMDIChild).Caption := lowerCase(SaveDialog.FileName);
            if (Pos('.BMP', Uppercase(SaveDialog.FileName)) <> 0)
            then TMDIChild(ActiveMDIChild).Bitmap.SaveToFile(SaveDialog.FileName);
            {$IFNDEF VER90}
              if (Pos('.JPG', Uppercase(SaveDialog.FileName)) <> 0)
              then begin
                   JPEGImage := TJPEGImage.Create;
                   with TMDIChild(ActiveMDIChild)
                   do begin
                      JPEGImage.Assign(Bitmap);
                      if (Bitmap.PixelFormat <> pf24bit) 
                      then JPEGImage.PixelFormat := jf24bit;
                   end;
                   JPEGImage.Scale := jsFullSize;
                   { Dont compress JPEG image, ie. set quality to 100.           }
                   JPEGImage.CompressionQuality  := 100;
                   JPEGImage.ProgressiveEncoding := True;
                   JPEGImage.SaveToFile(SaveDialog.FileName);
                   JPEGImage.Free;
              end;
            {$ENDIF}
            {$IFDEF INCTIFF}
              if (Pos('.TIF', SaveDialog.FileName) <> 0)
              then SaveTIFFToBitmap(PChar(SaveDialog.FileName), TMDIChild(ActiveMDIChild).Bitmap);
            {$ENDIF}
       end;
  end;
end; { End TMainForm.FileSaveItemClick.                                        }


procedure TMainForm.FileSaveAsItemClick(Sender : TObject);
begin
  { save current file under new name }
end; { End TMainForm.FileSaveAsItemClick.                                      }


procedure TMainForm.FileExitItemClick(Sender : TObject);
begin
  Close;
end; { End TMainForm.FileExitItemClick.                                        }


procedure TMainForm.CutItemClick(Sender : TObject);
begin
  {cut selection to clipboard}
end; { End TMainForm.CutItemClick.                                             }


procedure TMainForm.CopyItemClick(Sender : TObject);
begin
  {copy selection to clipboard}
end; { End TMainForm.CopyItemClick.                                            }


procedure TMainForm.PasteItemClick(Sender : TObject);
begin
  {paste from clipboard}
end; { End TMainForm.PasteItemClick.                                           }


procedure TMainForm.WindowCascadeItemClick(Sender : TObject);
begin
  Cascade;
end; { End TMainForm.WindowCascadeItemClick.                                   }


procedure TMainForm.WindowTileItemClick(Sender : TObject);
begin
  Tile;
end; { End TMainForm.WindowTileItemClick.                                      }


procedure TMainForm.WindowArrangeItemClick(Sender : TObject);
begin
  ArrangeIcons;
end; { End TMainForm.WindowArrangeItemClick.                                   }


procedure TMainForm.WindowMinimizeItemClick(Sender : TObject);
var i : integer;
begin
  { Must be done backwards through the MDIChildren array }
  for i := MDIChildCount - 1 downto 0
  do MDIChildren[i].WindowState := wsMinimized;
end; { End TMainForm.WindowMinimizeItemClick.                                  }


procedure TMainForm.UpdateMenuItems(Sender : TObject);
var EnableMenu : boolean;
begin
  EnableMenu                 := MDIChildCount > 0;
  FileSaveItem.Enabled       := EnableMenu;
  CutItem.Enabled            := EnableMenu;
  CopyItem.Enabled           := EnableMenu;
  PasteItem.Enabled          := EnableMenu;
  SaveBtn.Enabled            := EnableMenu;
  CutBtn.Enabled             := EnableMenu;
  CopyBtn.Enabled            := EnableMenu;
  PasteBtn.Enabled           := EnableMenu;
  WindowCascadeItem.Enabled  := EnableMenu;
  WindowTileItem.Enabled     := EnableMenu;
  WindowArrangeItem.Enabled  := EnableMenu;
  WindowMinimizeItem.Enabled := EnableMenu;
end; { End TMainForm.UpdateMenuItems.                                          }


procedure TMainForm.FormDestroy(Sender : TObject);
begin
  { TWAIN                                                                      }
  TWTerminate(0);
  Screen.OnActiveFormChange := nil;
end; { End TMainForm.FormDestroy.                                              }


procedure TMainForm.TWAIN_MessageHook;
var Msg : TMsg;
begin
  while GetMessage(msg, 0, 0, 0) and TWIsDSOpen
  do begin
     if Not(ProcessTWMessage(PMSG(@Msg), Handle))
     then begin
          TranslateMessage(Msg);
          DispatchMessage(Msg);
     end;
  end;
end; { End TMainForm..TWAIN_MessageHook.                                       }


{------------------------------------------------------------------------------}
{ This is where the image is returned from the TWAIN driver.                   }
{------------------------------------------------------------------------------}

procedure TMainForm.PMXferDone(var Msg : TMessage);
type LongType       = record
                        case Word of
                        0 : (Ptr  : Pointer);
                        1 : (Long : Longint);
                        2 : (Lo   : Word;
                             Hi   : Word);
                      end;
ta = array[0..8000] of byte;
pa = ^ta;
var OldCursor      : hCursor;
    ScreenDC       : HDC;

    Count          : Longint;
    StartS         : LongType;
    ToAddrS        : LongType;
    BitsS          : LongType;
    StartT         : LongType;
    ToAddrT        : LongType;
    BitsT          : LongType;

    hTmpBmp        : THandle;

    MemFlags       : longint;
    hMem           : THandle;
    hBmp           : THandle;
    hBmpInfo       : THandle;
    hWinBmp        : HBitmap;

    pMem           : pointer;
    pBmp           : pointer;
    pBmpInfo       : PBitmapInfo;
    pBmpInfoHeader : pBitmapInfoHeader;
    pBmpih         : pBitmapInfoHeader;
    pBmpch         : pBitmapCoreHeader;

    PalSize        : longint;
    RGBPalOfs      : longint;
    NoBits         : word;
    HeaderSize     : longint;
    SizeImage      : longint;
    LongWidth      : longint;

    BmpWidth       : longint;
    BmpHeight      : longint;

    ErrorStr       : string;
    a : pa;
begin
  OldCursor := SetCursor(LoadCursor(0, idc_Wait));

  {$IFDEF TWAIN32}
    LogMessage('PM_XFERDONE');
  {$ENDIF}

  { TWAIN                                                                      }
  { Scanned image was transfered through memory.                               }
  { NOTE : This version always provides images in memory.                      }
  if (Msg.wParam <> 0)
  then begin
       hBmp      := 0;
       hBmpinfo  := 0;
       BmpWidth  := 0;
       BmpHeight := 0;
       hMem      := THandle(Msg.wParam);

       { Convert the hMem handle to hBmp & hBmpInfo.                           }
       pMem := GlobalLock(hMem);
       if (pMem <> Nil)
       then begin
            PalSize        := 0;
            pBmp           := pMem;
            pBmpInfoHeader := pBmp;

            { Get palette size.                                                }
            pBmpih := pBitmapInfoHeader(pBmpInfoHeader);
            pBmpch := pBitmapCoreHeader(pBmpInfoHeader);

            { Get number of bit per pixel.                                     }
            if (pBmpih^.biSize <> SizeOf(TBitmapCoreHeader))
            then NoBits := pBmpih^.biBitCount * pBmpih^.biPlanes
            else NoBits := pBmpch^.bcBitCount;

            { Determind Palette size.                                          }
            RGBPalOfs := 0;
            case NoBits of
            1 :  PalSize := 2;
            4 :  PalSize := 16;
            8 :  PalSize := 256;
            24 : begin
                   PalSize := 0;
                   if (pBmpInfoHeader^.biClrUsed > 0)
                   then RGBPalOfs := 4 * pBmpInfoHeader^.biClrUsed;
                   pBmpInfoHeader^.biClrUsed := 0;
                 end;
            else PalSize := 0;
            end;

            { Calc. Btimap header size.                                        }
            HeaderSize := SizeOf(TBitmapInfoHeader) + PalSize * SizeOf(TRGBQuad);

            { Create & lock Bitmap Info.                                       }
            hBmpInfo   := GlobalAlloc(gmem_Moveable or gmem_Zeroinit, HeaderSize);
            pBmpInfo   := GlobalLock(hBmpInfo);

            { Copy Bitmap info.                                                }
            {$IFDEF WIN32}
              CopyMemory(pBmpInfo, pBmp, HeaderSize);
            {$ELSE}
              hmemcpy(pBmpInfo, pBmp, HeaderSize);
            {$ENDIF}

            with pBmpInfo^.bmiHeader
            do begin
               biClrUsed := 0;

               { Ensure proper information.                                    }
               if (biXPelsPerMeter <= 0)
               then if (biYPelsPerMeter > 0)
                    then biXPelsPerMeter := biYPelsPerMeter
                    else biXPelsPerMeter := round(30000.0 / 2.54);
               if (biYPelsPerMeter <= 0)
               then if (biXPelsPerMeter > 0)
                    then biYPelsPerMeter := biXPelsPerMeter
                    else biYPelsPerMeter := round(30000.0 / 2.54);

               { Calc. Long Width of Line.                                     }
               LongWidth := (((biWidth * biBitCount) + 31) div 32) * 4;
               if (biSizeImage = 0)
               then biSizeImage := LongWidth * biHeight;
               SizeImage := biSizeImage;
            end;

            { Set window size indirectly.                                      }
            BmpWidth  := pBmpInfo^.bmiHeader.biWidth;
            BmpHeight := pBmpInfo^.bmiHeader.biHeight;

            { Unlock Bitmap Info.                                              }
            GlobalUnlock(hBmpInfo);

            { Move bitmap to top of memory array.                              }
            Count       := SizeImage;
            StartS.Long := 0;
            StartT.Long := 0;
            BitsS.Ptr   := pMem;
            BitsT.Ptr   := pBmp;
            inc(StartS.Long, HeaderSize);
            inc(StartS.Long, RGBPalOfs);

            if (BitsT.Ptr <> nil) and (BitsS.Ptr <> nil)
            then begin
                 while (Count > 0)
                 do begin
                    { Move one line at a time.                                 }
                    if (Count  > LongWidth)
                    then Count := LongWidth;
                    {$IFDEF WIN32}
                      { 32 Bit code.                                           }
                      ToAddrS.Long := BitsS.Long + StartS.Long;
                      ToAddrT.Long := BitsT.Long + StartT.Long;
                      CopyMemory(ToAddrT.Ptr, ToAddrS.Ptr, Count);

                    {$ELSE}
                      { 16 Bit code.                                           }
                      ToAddrS.Hi := BitsS.Hi + (StartS.Hi * Ofs(AHIncr));
                      ToAddrS.Lo := StartS.Lo;
                      ToAddrT.Hi := BitsT.Hi + (StartT.Hi * Ofs(AHIncr));
                      ToAddrT.Lo := StartT.Lo;

                      { Move bytes in picture.                                 }
                      hmemcpy(ToAddrT.Ptr, ToAddrS.Ptr, Count);
                    {$ENDIF}
                    StartS.Long := StartS.Long + Count;
                    StartT.Long := StartT.Long + Count;
                    Count       := SizeImage - StartT.Long;
                 end;
            end;

            { Unlock returned bitmap handle.                                   }
            GlobalUnlock(hMem);

            { Down size the hMem to only hold the actual pixels in the bitmap. }
            MemFlags := GlobalFlags(hMem);
            hTmpBmp  := GlobalReAlloc(hMem, SizeImage + 16, LoWord(MemFlags));
            if (hTmpBmp <> 0)
            then begin
                 hBmp := hTmpBmp;
            end
            else begin
                 ErrorStr := Format('hMem resize error: %d', [GetLastError]);
                 MessageDlg(ErrorStr, mtInformation, [mbOK], 0);
                 hBmp := hMem;
            end;

            { If one of the memory handles doesn't contain a proper handle     }
            { then discharge it all.                                           }
            if (hBmp     = 0) or
               (hBmpInfo = 0)
            then begin
                 if (hBmp <> 0)
                 then GlobalFree(hBmp);
                 if (hBmpInfo <> 0)
                 then GlobalFree(hBmpInfo);
                 hBmp     := 0;
                 hBmpInfo := 0;
            end;
       end
       else begin
            GlobalFree(hMem);
            hMem := 0;
       end;

       { Create a MDI window.                                                  }
       if (hBmp <> 0) and (hBmpInfo <> 0)
       then begin
            { Create a DIBitmap that the canvas can show.                      }
            ScreenDC := GetDC(0);
            pBmp     := GlobalLock(hBmp);
            pBmpInfo := GlobalLock(hBmpInfo);
            hWinBmp  := CreateDIBitmap(ScreenDC,
                                       pBmpInfo^.bmiHeader,
                                       CBM_INIT,
                                       pBmp,
                                       pBmpInfo^,
                                       DIB_RGB_Colors);
            a := globallock(hWinBmp);
            globalunlock(hWinBmp);
            GlobalUnlock(hBmp);
            GlobalUnlock(hBmpInfo);
            WinProcs.ReleaseDC(0, ScreenDC);
            CreateMDIChild('', hWinBmp, BmpWidth, BmpHeight);
       end;

       if (hBmp <> 0)
       then GlobalFree(hBmp);
       if (hBmpInfo <> 0)
       then GlobalFree(hBmpInfo);

       Msg.wParam := 0;
  end;

  SetCursor(OldCursor);
  Msg.Result := 0;
end; { End TMainForm.PMXferDone.                                               }


procedure TMainForm.AcquireClick(Sender : TObject);
var OldCursor  : hCursor;
    Flag       : integer;
    ShowUI     : bool;
    TransType  : integer;
begin
  OldCursor := SetCursor(LoadCursor(0, idc_Wait));

  { Flag = 0 Do Not Accept MSG_XFERREADY           }
  {        1 Disable/CloseDS/CloseDSM              }
  {        2 Disable Only                          }
  {        3 Do Not Disable - only if ShowUI=True  }
  if FileShowUI.Checked
  then begin
       ShowUI := True;
       if FileDisable.Checked
       then Flag := 1
       else Flag := 3;
  end
  else begin
       ShowUI := False;
       Flag   := 2;
  end;
  if MemoryXfer.Checked
  then TransType := 1
  else if FileXfer.Checked
       then TransType := 2
       else TransType := 0;
  TWAcquire(Handle, ShowUI, Flag, TransType);
  TWAIN_MessageHook;
  SetCursor(OldCursor);
end; { End TMainForm.AcquireClick.                                             }


procedure TMainForm.SourceClick(Sender : TObject);
var OldCursor  : hCursor;
begin
  OldCursor := SetCursor(LoadCursor(0, idc_Wait));

  { TWAIN                                                                      }
  { SM userselect, or standard SM based user interface to Source's             }
  { - open dsm                                                                 }
  { - show user the data sources found by the SM                               }
  {   allow user to pick a particular Source                                   }
  { - close the SM                                                             }
  if TWOpenDSM
  then begin
       TWSelectDS;
       TWCloseDSM(0);
  end;
  SetCursor(OldCursor);
end; { End TMainForm.SourceClick.                                              }


procedure TMainForm.NativeXferClick(Sender : TObject);
begin
  NativeXfer.Checked        := False;
  FileXfer.Checked          := False;
  MemoryXfer.Checked        := False;
  TMenuItem(Sender).Checked := True;
end; { End TMainForm.NativeXferClick.                                          }


procedure TMainForm.FileXferClick(Sender : TObject);
begin
  NativeXfer.Checked        := False;
  FileXfer.Checked          := False;
  MemoryXfer.Checked        := False;
  TMenuItem(Sender).Checked := True;
end; { End TMainForm.FileXferClick.                                            }


procedure TMainForm.MemoryXferClick(Sender : TObject);
begin
  NativeXfer.Checked        := False;
  FileXfer.Checked          := False;
  MemoryXfer.Checked        := False;
  TMenuItem(Sender).Checked := True;
end; { End TMainForm.MemoryXferClick.                                          }


procedure TMainForm.MsgNoneClick(Sender : TObject);
begin
  MsgNone.Checked           := False;
  MsgError.Checked          := False;
  MsgInformation.Checked    := False;
  MsgFull.Checked           := False;
  TMenuItem(Sender).Checked := True;
  SetMessageLevel(ML_NONE);
end; { End TMainForm.MsgNoneClick.                                             }


procedure TMainForm.MsgErrorClick(Sender : TObject);
begin
  MsgNone.Checked           := False;
  MsgError.Checked          := False;
  MsgInformation.Checked    := False;
  MsgFull.Checked           := False;
  TMenuItem(Sender).Checked := True;
  SetMessageLevel(ML_ERROR);
end; { End TMainForm.MsgErrorClick.                                            }


procedure TMainForm.MsgInformationClick(Sender : TObject);
begin
  MsgNone.Checked           := False;
  MsgError.Checked          := False;
  MsgInformation.Checked    := False;
  MsgFull.Checked           := False;
  TMenuItem(Sender).Checked := True;
  SetMessageLevel(ML_INFO);
end; { End TMainForm.MsgInformationClick.                                      }


procedure TMainForm.MsgFullClick(Sender : TObject);
begin
  MsgNone.Checked           := False;
  MsgError.Checked          := False;
  MsgInformation.Checked    := False;
  MsgFull.Checked           := False;
  TMenuItem(Sender).Checked := True;
  SetMessageLevel(ML_FULL);
end; { End TMainForm.MsgFullClick.                                             }


procedure TMainForm.AutomaticFeedingClick(Sender : TObject);
begin
  if TMenuItem(Sender).Checked
  then TMenuItem(Sender).Checked := False
  else TMenuItem(Sender).Checked := True;
end; { End TMainForm.AutomaticFeedingClick.                                    }


procedure TMainForm.FileShowUIClick(Sender : TObject);
begin
  if TMenuItem(Sender).Checked
  then TMenuItem(Sender).Checked := False
  else TMenuItem(Sender).Checked := True;
end; { End TMainForm.FileShowUIClick.                                          }


procedure TMainForm.FileDisableClick(Sender : TObject);
begin
  if TMenuItem(Sender).Checked
  then TMenuItem(Sender).Checked := False
  else TMenuItem(Sender).Checked := True;
end; { End TMainForm.FileDisableClick.                                         }


procedure TMainForm.HelpAboutItemClick(Sender : TObject);
begin
  FormAbout := TFormAbout.Create(Application);
  FormAbout.ShowModal;
  FormAbout.Close;
end; { End TMainForm.HelpAboutItemClick.                                       }

end. { End Main.                                                               }

